// This file is part of VKF, a CPython library
// for Machine Learning based on Lattice Theory.
//
// This Source Code Form is subject to the terms of the X11 (MIT) License.
// If a copy of the X11 License was not distributed with this file, 
// You can obtain one at https://opensource.org/licenses/MIT/.
//
// X11 (MIT) License
// 
// Copyright (c) 2020 Dmitry V. Vinogradov
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef VKF_VERGES_HPP
#define VKF_VERGES_HPP

#include "Inhabit.hpp"
#include <boost/dynamic_bitset.hpp>

namespace vkf {

class Verges : public std::list<double> {
    private:
        double maximum;
        double minimum;
        Inhabit axe;
    public:
        double GetMaximum() const { return maximum; };
        double GetMinimum() const { return minimum; };
        ~Verges() = default;
    Verges(Inhabit splitter) : axe(splitter) {
        maximum = axe.back().first;
        minimum = axe.front().first;
        // for (auto db_pair : axe) {
        //     if (db_pair.first > maximum) {
        //         maximum = db_pair.first;
        //     }
        //     if (db_pair.first < minimum) {
        //         minimum = db_pair.first;
        //     }
        // }
    };
    void Populate (std::size_t length) {
        if (length == 0) {
            return;
        }
        double current_entropy;
        double max_entropy = 0.0;
        std::list<std::pair<double,bool>>::iterator begin_it = axe.begin();
        std::list<std::pair<double,bool>>::iterator end_it = axe.end();
//        for (auto it = begin(); it != end(); it++) {
//            stop_it = std::find_if(axe.begin(),axe.end(),
//                [&](const std::pair<double,bool> &arg){ return arg.first == (*it); });
//            current_entropy = axe.ComputeEntropy(start_it, stop_it);
//            if (max_entropy < current_entropy) {
//                max_entropy = current_entropy;
//                begin_it = start_it;
//                end_it = stop_it;
//            }
//            start_it = stop_it;
//        }
//        current_entropy = axe.ComputeEntropy(start_it, axe.end());
//        if (max_entropy <= current_entropy) {
//            max_entropy = current_entropy;
//            begin_it = start_it;
//            end_it = axe.end();
//        }
        bool robust;
        std::list<std::pair<double,bool>>::iterator current_it;
        std::list<std::pair<double,bool>>::iterator medium_it;
        std::list<std::pair<double,bool>>::iterator previous_it = axe.begin();
        for (auto it = this->begin(); it != this->end(); it++) {
            current_it = std::find_if(axe.begin(),axe.end(),
                [&](const std::pair<double,bool> &arg){ return arg.first == (*it); });
            medium_it = std::find_if(axe.begin(),axe.end(),
                [&](const std::pair<double,bool> &arg){ return ((arg.first > previous_it->first) && (arg.first < current_it->first)); });
            robust = medium_it != axe.end()? false : true;
            if (!robust) {
                current_entropy = axe.ComputeEntropy(previous_it, current_it);
                if (max_entropy < current_entropy) {
                    max_entropy = current_entropy;
                    begin_it = previous_it;
                    end_it = current_it;
                }
            }
            previous_it = current_it;
        }
        medium_it = std::find_if(axe.begin(),axe.end(),
            [&](const std::pair<double,bool> &arg){ return ((arg.first > previous_it->first) && (arg.first < this->maximum)); });
        robust = (medium_it != axe.end()? false : true);
        if (!robust) {
            current_entropy = axe.ComputeEntropy(previous_it, axe.end());
            if (max_entropy <= current_entropy) {
                max_entropy = current_entropy;
                begin_it = previous_it;
                end_it = axe.end();
            }
        }
        if (max_entropy > 0.0) {
            double middle_verge = axe.ComputeVerge(begin_it, end_it);
            medium_it = std::find_if(axe.begin(),axe.end(),
                [&](const std::pair<double,bool> &arg){ return arg.first == middle_verge; });
            if(medium_it == axe.end()) {
                return;
            } else {
                this->push_back(middle_verge);
            }
            sort();
            if (length > 1) {
                Populate(length-1);
            }
            return;
        } else {
            return;
        }
    }
};

} // namespace vkf

#endif // VKF_VERGES_HPP
